from .reference import ReferenceFileModel

__all__ = ['MirMrsPtCorrModel']


class MirMrsPtCorrModel(ReferenceFileModel):
    """
    A data model for MIRI mrs IFU across-slice corrections file.

    Parameters
    __________
    init : any
        Any of the initializers supported by `~jwst.datamodels.DataModel`.

    leakcor_table : numpy table
         IFU spectral leak correction (fractional, Jy to Jy)

    tracor_table : numpy table
         IFU across slice transmission correction

    wavcorr_optical_table : numpy table
         IFU across slice wavelength offset table 1

    wavcorr_xslice_table : numpy table
         IFU across slice wavelength offset table 2

    wavcorr_shift_table : numpy table
         IFU across slice wavelength offset table 3

    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/miri_mrsptcorr.schema"



"""
Various utility functions and data types
"""

import copy
import os

import numpy as np
from numpy.lib.recfunctions import merge_arrays
from astropy.io import fits
import asdf
from asdf import treeutil

try:
    from asdf.treeutil import RemoveNode
except ImportError:
    # Prior to asdf 2.8, None was used to indicate
    # that a node should be removed.
    RemoveNode = None

import logging
log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


def gentle_asarray(a, dtype, allow_extra_columns=False):
    """
    Convert ``a`` to dtype ``dtype`` ignoring case differences in
    dtype field (column) names for structured arrays (tables).

    When name conflicts occur (the cases differ) the name from
    ``dtype`` will be used.

    Parameters
    ----------

    a : np.ndarray
        Array which will be converted to the new dtype.

    dtype : np.dtype
        The dtype of the new array.

    Returns
    -------

    new_array : np.ndarray
        Array converted to the new dtype.
    """
    if isinstance(dtype, np.dtype):
        out_dtype = copy.copy(dtype)
    else:
        out_dtype = np.dtype(dtype)

    if not isinstance(a, np.ndarray):
        try:
            a = np.asarray(a, dtype=out_dtype)
        except Exception:
            raise ValueError("Can't convert {0!s} to ndarray".format(type(a)))
        return a
    in_dtype = a.dtype

    # Non-table array
    if in_dtype.fields is None and out_dtype.fields is None:
        if np.can_cast(in_dtype, out_dtype, 'equiv'):
            return a
        else:
            return _safe_asanyarray(a, out_dtype)

    # one of these dtypes does not have fields
    if in_dtype.fields is None or out_dtype.fields is None:
        return _safe_asanyarray(a, out_dtype)

    # this function should not handle nested structured dtypes
    # as they aren't supported by FITS_rec and not handled well
    # by merge_arrays below (which currently flattens the dtype)
    for dt in (in_dtype, out_dtype):
        for n in dt.names:
            if dt[n].names is not None:
                msg = f"gentle_asarray does not support nested structured dtypes: {dt}"
                raise ValueError(msg)

    # When a FITS file includes a pseudo-unsigned-int column, astropy will return
    # a FITS_rec with an incorrect table dtype.  The following code rebuilds
    # in_dtype from the individual fields, which are correctly labeled with an
    # unsigned int dtype.
    # We can remove this once the issue is resolved in astropy:
    # https://github.com/astropy/astropy/issues/8862
    if isinstance(a, fits.fitsrec.FITS_rec):
        a.dtype = rebuild_fits_rec_dtype(a)
        in_dtype = a.dtype

    if in_dtype == out_dtype:
        return a

    # check if names match (ignoring case)
    in_lower_names = [n.lower() for n in in_dtype.names]
    out_lower_names = [n.lower() for n in out_dtype.names]
    if in_lower_names == out_lower_names:
        in_subdtypes = [in_dtype[n] for n in in_dtype.names]
        out_subdtypes = [out_dtype[n] for n in out_dtype.names]
        if in_subdtypes == out_subdtypes:
            return a.view(dtype=out_dtype)
        else:
            # else, use asanyarray and copy
            return _safe_asanyarray(a, out_dtype)

    # names don't match
    # check if names match but the order is incorrect
    if set(out_lower_names) == set(in_lower_names):
        # all the columns exist but they are in the wrong order
        # reorder the columns, the names might differ in case
        reordered_names = sorted(in_dtype.names, key=lambda n: out_lower_names.index(n.lower()))
        reordered_array = merge_arrays([a[n] for n in reordered_names], flatten=True)
        reordered_subdtypes = [reordered_array.dtype[n] for n in reordered_array.dtype.names]
        out_subdtypes = [out_dtype[n] for n in out_dtype.names]
        if reordered_subdtypes == out_subdtypes:
            return reordered_array.view(out_dtype)
        else:
            return _safe_asanyarray(reordered_array, out_dtype)

    # if extra columns are not allowed or they are (and the required columns are missing)
    # then raise an exception
    if not allow_extra_columns or (not set(out_lower_names).issubset(in_lower_names)):
        # try to match the old error message
        raise ValueError(
            "Column names don't match schema. "
            "Schema has {0}. Data has {1}".format(
                str(set(out_lower_names).difference(set(in_lower_names))),
                str(set(in_lower_names).difference(set(out_lower_names)))))

    # construct new dtype with required columns at start
    # in_dtype vs out_dtype
    # - might have inconsequential (since fitsrec is used) name differences
    # - might have dtype differences (requiring _safe_asanyarray)
    # - in_dtype will have more columns (since we've handled everything else above)
    # if the first set of columns match (by name and dtype) the required dtype, return the array
    n_required = len(out_dtype)
    if in_dtype.descr[:n_required] == out_dtype.descr:
        return a

    # if the names of the first n_required columns match the columns
    if in_lower_names[:n_required] == out_lower_names:
        in_subdtypes = [in_dtype[n] for n in in_dtype.names]
        out_subdtypes = [out_dtype[n] for n in out_dtype.names]
        new_dtype = copy.copy(in_dtype)
        new_dtype.names = tuple(out_dtype.names + in_dtype.names[n_required:])
        if in_subdtypes[:n_required] == out_subdtypes:
            return a.view(dtype=new_dtype)
        else:
            new_dtype = np.dtype(out_dtype.descr + new_dtype.descr[len(out_dtype.descr):])
            return _safe_asanyarray(a, new_dtype)

    # reorder columns so required columns are first
    required_names = [n for n in in_dtype.names if n.lower() in out_lower_names]
    required_names.sort(key=lambda n: out_lower_names.index(n.lower()))
    extra_names = [n for n in in_dtype.names if n.lower() not in out_lower_names]
    names_ordered = tuple(required_names + extra_names)
    reordered_array = merge_arrays([a[n] for n in names_ordered], flatten=True)
    reordered_array.dtype.names = names_ordered

    extra_dtype_descr = [(n, in_dtype[n]) for n in extra_names]
    new_dtype = np.dtype(out_dtype.descr + extra_dtype_descr)

    # check that required columns have the correct dtype
    reordered_subdtypes = [reordered_array.dtype[n] for n in reordered_array.dtype.names]
    out_subdtypes = [out_dtype[n] for n in out_dtype.names]
    if reordered_subdtypes[:n_required] == out_subdtypes:
        return reordered_array.view(new_dtype)
    return _safe_asanyarray(reordered_array, new_dtype)


def _safe_asanyarray(a, dtype):
    if isinstance(a, fits.fitsrec.FITS_rec):
        if any(c.bzero is not None for c in a.columns):
            # Due to an issue in astropy, it's not safe to directly cast
            # a FITS_rec with a pseudo-unsigned column.
            # See https://github.com/astropy/astropy/issues/12112
            result = np.zeros(a.shape, dtype=dtype)
            for old_col, new_col in zip(a.dtype.names, result.dtype.names):
                result[new_col] = a[old_col]
            return result

    result = np.asanyarray(a, dtype=dtype)
    if isinstance(result, fits.fitsrec.FITS_rec) and isinstance(a, fits.fitsrec.FITS_rec):
        for column in result.columns:
            name = column.name
            try:
                matching_column = a.columns[name]
            except KeyError:
                continue
            result.columns[name].unit = matching_column.unit
    return result


def create_history_entry(description, software=None):
    """
    Create a HistoryEntry object.

    Parameters
    ----------
    description : str
        Description of the change.
    software : dict or list of dict
        A description of the software used.  It should not include
        asdf itself, as that is automatically notated in the
        `asdf_library` entry.

        Each dict must have the following keys:

        ``name``: The name of the software
        ``author``: The author or institution that produced the software
        ``homepage``: A URI to the homepage of the software
        ``version``: The version of the software

    Examples
    --------
    >>> soft = {'name': 'jwreftools', 'author': 'STSCI', \
                'homepage': 'https://github.com/spacetelescope/jwreftools', 'version': "0.7"}
    >>> entry = create_history_entry(description="HISTORY of this file", software=soft)

    """
    from asdf.tags.core import Software, HistoryEntry
    import datetime

    if isinstance(software, list):
        software = [Software(x) for x in software]
    elif software is not None:
        software = Software(software)

    dt = datetime.datetime.now(datetime.timezone.utc).replace(tzinfo=None)
    entry = HistoryEntry({
        'description': description,
        'time': dt
    })

    if software is not None:
        entry['software'] = software
    return entry


def get_envar_as_boolean(name, default=False):
    """Interpret an environmental as a boolean flag

    Truth is any numeric value that is not 0 or
    any of the following case-insensitive strings:

    ('true', 't', 'yes', 'y')

    Parameters
    ----------
    name : str
        The name of the environmental variable to retrieve

    default : bool
        If the environmental variable cannot be accessed, use as the default.
    """
    truths = ('true', 't', 'yes', 'y')
    falses = ('false', 'f', 'no', 'n')
    if name in os.environ:
        value = os.environ[name]
        try:
            value = bool(int(value))
        except ValueError:
            value_lowcase = value.lower()
            if value_lowcase not in truths + falses:
                raise ValueError(f'Cannot convert value "{value}" to boolean unambiguously.')
            return value_lowcase in truths
        return value

    log.debug(f'Environmental "{name}" cannot be found. Using default value of "{default}".')
    return default


def get_model_type(init):
    """
    Fetch the model type string from the underlying file object.

    Parameters
    ----------
    init : asdf.AsdfFile or astropy.io.fits.HDUList

    Returns
    -------
    str or None
    """
    if isinstance(init, asdf.AsdfFile):
        if "meta" in init:
            return init["meta"].get("model_type")
        else:
            return None
    elif isinstance(init, fits.HDUList):
        return init[0].header.get("DATAMODL")
    else:
        raise TypeError(f"Unhandled init type: {init.__class__.__name__}")


def remove_none_from_tree(tree):
    """
    Remove None values from a tree.  Both dictionary keys
    and list indices with None values will be removed.

    Parameters
    ----------
    tree : object
        The root node of the tree.

    Returns
    -------
    object
        Modified tree.
    """

    def _remove_none(node):
        if node is None:
            return RemoveNode
        else:
            return node

    return treeutil.walk_and_modify(tree, _remove_none)


def convert_fitsrec_to_array_in_tree(tree):
    def _convert_fitsrec(node):
        if isinstance(node, fits.FITS_rec):
            return _fits_rec_to_array(node)
        else:
            return node
    return treeutil.walk_and_modify(tree, _convert_fitsrec)


def rebuild_fits_rec_dtype(fits_rec):
    dtype = fits_rec.dtype
    new_dtype = []
    for field_name in dtype.fields:
        table_dtype = dtype[field_name]
        shape = table_dtype.shape
        if shape:
            table_dtype = table_dtype.base
        field_dtype = fits_rec.field(field_name).dtype
        if np.issubdtype(table_dtype, np.signedinteger) and np.issubdtype(field_dtype, np.unsignedinteger):
            new_dtype.append((field_name, field_dtype, shape))
        else:
            new_dtype.append((field_name, table_dtype, shape))
    return np.dtype((np.record, new_dtype))


def _fits_rec_to_array(fits_rec):
    bad_columns = [n for n in fits_rec.dtype.fields if np.issubdtype(fits_rec[n].dtype, np.unsignedinteger)]
    if not len(bad_columns):
        return fits_rec.view(np.ndarray)
    new_dtype = rebuild_fits_rec_dtype(fits_rec)
    arr = np.asarray(fits_rec, new_dtype).copy()
    for name in bad_columns:
        arr[name] = fits_rec[name]
    return arr



from .reference import ReferenceFileModel


__all__ = ['NRMModel']


class NRMModel(ReferenceFileModel):
    """
    A data model for Non-Redundant Mask.

    Parameters
    __________
    nrm : numpy float32 array
         Non-Redundant Mask
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/nrm.schema"



from asdf.extension import ManifestExtension

from .converters.jwst_models import (Gwa2SlitConverter, Slit2MsaConverter, LogicalConverter,
                                     NirissSOSSConverter, RefractionIndexConverter,
                                     MIRI_AB2SliceConverter, NIRCAMGrismDispersionConverter,
                                     NIRISSGrismDispersionConverter, GratingEquationConverter,
                                     SnellConverter, Rotation3DToGWAConverter,
                                     CoordsConverter, V23ToSkyConverter)


# The order here is important; asdf will prefer to use extensions
# that occur earlier in the list.
TRANSFORM_EXTENSIONS = [
    ManifestExtension.from_uri(
        "asdf://stsci.edu/jwst_pipeline/manifests/jwst_transforms-1.0.0",
        legacy_class_names=["jwst.transforms.jwextension.JWSTExtension"],
        converters=[
            CoordsConverter(),
            Gwa2SlitConverter(),
            Slit2MsaConverter(),
            LogicalConverter(),
            NirissSOSSConverter(),
            RefractionIndexConverter(),
            Rotation3DToGWAConverter(),
            MIRI_AB2SliceConverter(),
            NIRCAMGrismDispersionConverter(),
            NIRISSGrismDispersionConverter(),
            GratingEquationConverter(),
            SnellConverter(),
        ],
    ),
    ManifestExtension.from_uri(
        "asdf://stsci.edu/jwst_pipeline/manifests/jwst_transforms-0.7.0",
        legacy_class_names=["jwst.transforms.jwextension.JWSTExtension"],
        converters=[
            CoordsConverter(),
            Gwa2SlitConverter(),
            Slit2MsaConverter(),
            LogicalConverter(),
            NirissSOSSConverter(),
            RefractionIndexConverter(),
            Rotation3DToGWAConverter(),
            MIRI_AB2SliceConverter(),
            NIRCAMGrismDispersionConverter(),
            NIRISSGrismDispersionConverter(),
            GratingEquationConverter(),
            SnellConverter(),
            V23ToSkyConverter(),
        ],
    )
]



from .reference import ReferenceFileModel


__all__ = ['EmiModel']


class EmiModel(ReferenceFileModel):
    """
    A data model to correct MIRI images for EMI contamination.

    Parameters
    __________
    data : numpy table
        The reference waves to correct for MIRI EMI.
        A table-like object containing phase amplitude values
        corresponding to the appropriate frequency
        - Hz390: float32 1D array
        - Hz218a: float32 1D array
        - Hz218b: float32 1D array
        - Hz218c: float32 1D array
        - Hz164: float32 1D array
        - Hz10: float32 1D array
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/emi.schema"
    reftype = "emicorr"

    def __init__(self, init=None, **kwargs):
        super(EmiModel, self).__init__(init=init, **kwargs)

    def on_save(self, path=None):
        self.meta.reftype = self.reftype
        self.meta.instrument.name = "MIRI"

    def validate(self):
        super(EmiModel, self).validate()



from datetime import datetime, timedelta, timezone

import pytest

import asdf
from astropy.io import fits
import numpy as np
from numpy.testing import assert_array_equal
from asdf.tags.core import Software, HistoryEntry

from stdatamodels import util


def test_gentle_asarray():
    x = np.array([('abc', 1.0)], dtype=[
        ('FOO', 'S3'),
        ('BAR', '>f8')])

    new_dtype = [('foo', '|S3'), ('bar', '<f8')]

    y = util.gentle_asarray(x, new_dtype)

    assert y['bar'][0] == 1.0


def test_gentle_asarray_array_input():
    """
    Test gentle_asarray with various forms of np.ndarray input,
    which may need to be cast to a different data type.
    """
    # Array that already bears the correct dtype:
    inp = np.array([1, 2, 3, 4], dtype=np.int16)
    result = util.gentle_asarray(inp, dtype=np.int16)
    assert result.dtype == np.int16
    assert_array_equal(result, inp)

    # This will require casting, since int8 != int16:
    result = util.gentle_asarray(inp, dtype=np.int8)
    assert result.dtype == np.int8
    assert_array_equal(result, inp)

    # Test broadcasting the same 1D array to a recarray with
    # two columns.  The input array should be copied into
    # each column and cast as necessary:
    dtype = np.dtype([("col1", np.int16), ("col2", np.float32)])
    result = util.gentle_asarray(inp, dtype=dtype)
    assert result.dtype == dtype
    assert result["col1"].dtype == np.int16
    assert_array_equal(result["col1"], np.array([1, 2, 3, 4], dtype=np.int16))
    assert result["col2"].dtype == np.float32
    assert_array_equal(result["col1"], np.array([1, 2, 3, 4], dtype=np.float32))


def test_gentle_asarray_recarray_input():
    """
    Test gentle_asarray input that is already an np.recarray.
    """
    # This is just the simple case where the dtype already matches
    # the desired output, but it follows a different code path
    # from array input:
    dtype = np.dtype([("col1", np.int16), ("col2", np.float32)])
    inp = np.array([1, 2, 3, 4], dtype=dtype)
    result = util.gentle_asarray(inp, dtype=dtype)
    assert result.dtype == dtype
    assert_array_equal(result, inp)


def test_gentle_asarray_fits_rec_input():
    """
    Test gentle_asarray with FITS_rec array input.
    """
    # 'e' is the FITS data type code for single-precision float:
    cols = [fits.Column("col1", format="e", array=np.array([1, 2, 3, 4]))]
    inp = fits.FITS_rec.from_columns(cols)
    out_dtype = np.dtype([("col1", np.float32)])
    result = util.gentle_asarray(inp, dtype=out_dtype)
    assert result.dtype == out_dtype
    assert_array_equal(result, inp)


def test_gentle_asarray_fits_rec_pseudo_unsigned(tmp_path):
    """
    Test gentle_asarray handling of a FITS_rec with a pseudo unsigned
    integer column, which is a special case due to a bug in astropy.
    """
    # 'j' is the FITS data type code for 32-bit integer.  A column
    # is only considered "pseudo unsigned" for certain values of
    # bezero, one of which is 2**31.
    cols = [fits.Column("col1", format="j", array=np.array([1, 2, 3, 4], np.uint32), bzero=2**31)]
    inp = fits.FITS_rec.from_columns(cols)
    out_dtype = np.dtype([("col1", np.uint32)])
    result = util.gentle_asarray(inp, dtype=out_dtype)
    # If not handled, the astropy bug would result in a signed
    # int column dtype:
    assert result["col1"].dtype == np.uint32
    assert result["col1"][0] == 1
    assert result["col1"][1] == 2
    assert result["col1"][2] == 3
    assert result["col1"][3] == 4

    # This tests the case where a table with a pseudo unsigned integer column
    # is opened from a FITS file and needs to be cast.  This requires special
    # handling on our end to dodge the bug.
    file_path = tmp_path / "test.fits"

    dtype = np.dtype('>u2')
    data = np.array([(0,)], dtype=[("col1", dtype)])
    hdu = fits.BinTableHDU()
    hdu.data = data
    hdul = fits.HDUList([fits.PrimaryHDU(), hdu])
    hdul.writeto(file_path)

    with fits.open(file_path) as hdul:
        for dtype_str in ('>u2', '<u2', '>i2', '<i2'):
            dtype = np.dtype(dtype_str)
            result = util.gentle_asarray(hdul[-1].data, dtype=[("col1", dtype)])
            # Without the fix, the value in the array would be 128 due to bzero
            # shift being applied twice.
            assert result[0][0] == 0


def test_gentle_asarray_fits_rec_pseudo_unsigned_shaped(tmp_path):
    # This tests the case where a table with a pseudo unsigned integer column
    # is opened from a FITS file and needs to be cast.  This requires special
    # handling on our end to dodge the bug.
    file_path = tmp_path / "test.fits"

    data = np.array([((1,2),), ((3,4),)], dtype=[("col1", '>u2', 2)])
    hdu = fits.BinTableHDU()
    hdu.data = data
    hdul = fits.HDUList([fits.PrimaryHDU(), hdu])
    hdul.writeto(file_path)

    with fits.open(file_path) as hdul:
        for sdtype in  (np.dtype('>i2'), np.dtype('>u2')):
            result = util.gentle_asarray(hdul[-1].data, dtype=[("col1", sdtype, 2)])
            assert result['col1'].dtype.base == sdtype
            assert result.dtype['col1'].base == sdtype
            assert result.dtype['col1'].shape == (2, )
            assert result['col1'][0][0] == 1
            assert result['col1'][0][1] == 2
            assert result['col1'][1][0] == 3
            assert result['col1'][1][1] == 4
            assert result['col1'][[True, False]][0][0] == 1
            assert result['col1'][[True, False]][0][1] == 2


def test_gentle_asarray_nested_array():
    """
    Test gentle_asarray with nested arrays in one of the record
    fields.
    """
    # '2f' is a numpy data type code for an array of two 32-bit floats:
    in_dtype = np.dtype([("col1", np.dtype("2f")), ("col2", np.int16)])
    inp = np.array([1, 2, 3, 4], dtype=in_dtype)
    out_dtype = np.dtype([("col1", np.dtype("2f")), ("col2", np.int8)])
    result = util.gentle_asarray(inp, dtype=out_dtype)
    assert result.dtype == out_dtype
    assert_array_equal(result["col2"], inp["col2"])


def test_gentle_asarray_mismatched_field_names():
    """
    Test gentle_asarray with input field names that don't
    match the requested output field names.
    """
    in_dtype = np.dtype([("col1", np.int16), ("col2", np.float32)])
    inp = np.array([1, 2, 3, 4], dtype=in_dtype)
    out_dtype = np.dtype([("col1", np.int16), ("foo", np.float32)])
    with pytest.raises(ValueError):
        util.gentle_asarray(inp, dtype=out_dtype)


def test_gentle_asarray_field_name_case():
    """
    Test gentle_assary with input and output field names that
    only differ by case.
    """
    in_dtype = np.dtype([("col1", np.int16), ("col2", np.float32)])
    inp = np.array([1, 2, 3, 4], dtype=in_dtype)
    out_dtype = np.dtype([("COL1", np.int16), ("COL2", np.float32)])
    result = util.gentle_asarray(inp, dtype=out_dtype)
    assert result.dtype == out_dtype


def test_gentle_asarray_scalar_input():
    """
    Test gentle_asarray with scalar input.
    """
    result = util.gentle_asarray(3.14159, dtype=np.float32)
    assert result.dtype == np.float32
    assert_array_equal(result, np.array([3.14159], dtype=np.float32))


def test_gentle_asarray_invalid_conversion():
    """
    Test gentle_asarray with input that cannot be cast
    to an array.
    """
    with pytest.raises(ValueError):
        util.gentle_asarray(object(), dtype=np.float32)


@pytest.mark.parametrize("reorder", [True, False], ids=['different_order', 'same_order'])
@pytest.mark.parametrize("change_dtype", [True, False], ids=['different_dtype', 'same_dtype'])
@pytest.mark.parametrize("extra_columns", [True, False], ids=['extra_columns', 'no_extra_columns'])
@pytest.mark.parametrize("allow_extra", [True, False], ids=['allow_extra', 'disallow_extra'])
@pytest.mark.parametrize("change_case", [True, False], ids=['changed_case', 'same_case'])
def test_gentle_asarray_structured_dtype_configurations(reorder, change_dtype, extra_columns, allow_extra, change_case):
    """
    Test gentle_asarray with a structured array with a few combinations of:
        - misordered columns
        - columns with different dtypes
        - extra columns
        - allowing extra columns
    """
    # start with a target dtype which should (if no error occurs) be the dtype of the result
    target_dtype = np.dtype([('i', 'i4'), ('f', 'f8'), ('s', 'S3'), ('b', 'bool'), ('u', 'uint8'), ('e', 'i4')])
    input_descr = target_dtype.descr
    if extra_columns:
        # add 3 extra columns
        input_descr.append(('e1', 'i4'))
        input_descr.append(('e2', 'f8'))
        input_descr.append(('e3', 'S4'))
    if change_dtype:
        # change the dtype of a few columns
        input_descr[0] = ('i', 'i8')
        input_descr[1] = ('f', 'f4')
        input_descr[5] = ('e', 'f8')
        if extra_columns:
            # if we have extra columns, change those as well
            input_descr[6] = ('e1', 'f4')
            input_descr[7] = ('e2', 'i4')
    if reorder:
        # swap 2 columns
        input_descr[3], input_descr[2] = input_descr[2], input_descr[3]
        if extra_columns:
            # and swap the first extra column with the last required column
            input_descr[5], input_descr[6] = input_descr[6], input_descr[5]

    # generate the input datatype and data
    input_dtype = np.dtype(input_descr)
    input_array = np.zeros(5, input_dtype)
    input_array['i'] = 2
    input_array['f'] = 0.1
    input_array['s'] = b'a'
    input_array['b'] = True
    input_array['u'] = 3
    if change_case:
        input_array.dtype.names = tuple([n.upper() for n in input_array.dtype.names])
        input_dtype = input_array.dtype
    if not allow_extra and extra_columns:
        # if we have extra columns, but don't allow them, gentle_asarray should fail
        with pytest.raises(ValueError):
            new_array = util.gentle_asarray(input_array, target_dtype, allow_extra_columns=allow_extra)
        return

    new_array = util.gentle_asarray(input_array, target_dtype, allow_extra_columns=allow_extra)
    # check data passed through correctly
    assert np.all(new_array['i'] == 2)
    assert np.allclose(new_array['f'], 0.1)
    assert np.all(new_array['s'] == b'a')
    assert np.all(new_array['b'] == True)
    assert np.all(new_array['u'] == 3)
    if not extra_columns:
        # if we have not extra columns, the output dtype should match the target
        assert new_array.dtype == target_dtype
    else:
        # if we have extra columns, the required columns should have a dtype
        # that matches the required dtype
        n_required = len(target_dtype.descr)
        assert new_array.dtype.descr[:n_required] == target_dtype.descr
        # for the extra columns, the dtype should match the input
        for extra_name in new_array.dtype.names[n_required:]:
            new_array.dtype[extra_name] == input_dtype[extra_name]



def test_gentle_asarray_nested_structured_dtype():
    dt = np.dtype([('a', 'f8'), ('b', [('sa', 'f4'), ('sb', 'i4')])])
    arr = np.zeros(3, dt)
    with pytest.raises(ValueError, match=r'.*nested structured dtypes'):
        util.gentle_asarray(arr, dt)


def test_create_history_entry():
    entry = util.create_history_entry("Once upon a time...")
    assert isinstance(entry, HistoryEntry)
    assert entry["description"] == "Once upon a time..."
    assert entry.get("software") is None
    dt = datetime.now(timezone.utc).replace(tzinfo=None)
    assert (dt - entry["time"]) < timedelta(seconds=10)

    software = {"name": "PolarBearSoft", "version": "1.2.3"}
    entry = util.create_history_entry("There was a tie-dyed polar bear...", software)
    assert isinstance(entry["software"], Software)
    assert entry["software"]["name"] == "PolarBearSoft"
    assert entry["software"]["version"] == "1.2.3"

    software = [
        {"name": "PolarBearSoft", "version": "1.2.3"},
        {"name": "BanjoSoft", "version": "4.5.6"},
    ]
    entry = util.create_history_entry("Who loved to strum the banjo.", software)
    assert isinstance(entry["software"], list)
    assert all(isinstance(s, Software) for s in entry["software"])
    assert entry["software"][0]["name"] == "PolarBearSoft"
    assert entry["software"][0]["version"] == "1.2.3"
    assert entry["software"][1]["name"] == "BanjoSoft"
    assert entry["software"][1]["version"] == "4.5.6"


@pytest.mark.parametrize(
    "value,expected_result",
    [
        ("0", False),
        ("false", False),
        ("FALSE", False),
        ("f", False),
        ("no", False),
        ("n", False),
        ("1", True),
        ("-1", True),
        ("198815238", True),
        ("true", True),
        ("TRUE", True),
        ("t", True),
        ("yes", True),
        ("y", True),
    ]
)
def test_get_envar_as_boolean(monkeypatch, value, expected_result):
    monkeypatch.setenv("TEST_VAR", value)
    assert util.get_envar_as_boolean("TEST_VAR") is expected_result


def test_get_envar_as_boolean_default(monkeypatch):
    monkeypatch.delenv("TEST_VAR", raising=False)
    assert util.get_envar_as_boolean("TEST_VAR") is False
    assert util.get_envar_as_boolean("TEST_VAR", default=False) is False
    assert util.get_envar_as_boolean("TEST_VAR", default=True) is True


def test_get_envar_as_boolean_invalid(monkeypatch):
    monkeypatch.setenv("TEST_VAR", "foo")
    with pytest.raises(ValueError):
        assert util.get_envar_as_boolean("TEST_VAR")


def test_get_model_type():
    assert util.get_model_type(asdf.AsdfFile()) is None
    assert util.get_model_type(asdf.AsdfFile({"meta": {"model_type": "SomeModel"}})) == "SomeModel"

    assert util.get_model_type(fits.HDUList([fits.PrimaryHDU()])) is None

    hdu = fits.PrimaryHDU()
    hdu.header["DATAMODL"] = "SomeOtherModel"
    assert util.get_model_type(fits.HDUList([hdu])) == "SomeOtherModel"



from .model_base import JwstDataModel


__all__ = ['AmiLgFitModel']


class AmiLgFitModel(JwstDataModel):
    """
    A data model for AMI LG analysis results.

    Parameters
    ----------

    centered_image : numpy float32 array
        Centered image
    norm_centered_image : numpy float32 array
        Centered image normalized by data peak
    fit_image : numpy float32 array
        Fitted image
    norm_fit_image : numpy float32 array
        Fitted image normalized by data peak
    resid_image : numpy float32 array
        Residual image
    norm_resid_image : numpy float32 array
        Residual image normalized by data peak
    solns_table : numpy table
        Solutions table
    """
    schema_url = "http://stsci.edu/schemas/jwst_datamodel/amilgfitmodel.schema"

    def get_primary_array_name(self):
        """
        Returns the name "primary" array for this model, which
        controls the size of other arrays that are implicitly created.
        This is intended to be overridden in the subclasses if the
        primary array's name is not "data".
        """
        return 'fit_image'



import numpy
import pytest

from stdatamodels.dynamicdq import dynamic_mask


def _dq(init):
    return numpy.array(init, dtype='u4')


def _dq_def(init):
    return numpy.array(init, dtype=[('VALUE', 'u4'), ('NAME', 'S40')])


class FakeModel:
    def __init__(self, dq, dq_def):
        self.dq = dq
        self.dq_def = dq_def


@pytest.mark.parametrize('dq', [[], [1, 2, 3]])
@pytest.mark.parametrize('dq_def', [None, 1, _dq_def([])])
@pytest.mark.parametrize('mmap', [{}, {'a': 1}])
@pytest.mark.parametrize('inv', [True, False])
def test_copy_dq(dq, dq_def, mmap, inv):
    dq = _dq(dq)
    model = FakeModel(dq, dq_def)
    dqmask = dynamic_mask(model, mmap, inv)
    assert dq is dqmask


@pytest.mark.parametrize('inv', [True, False])
def test_mmap(inv):
    dq = _dq([0, 1, 2, 4])
    dq_def = _dq_def([(1, 'FOO'), (2, 'BAR'), (4, 'BAM')])
    model = FakeModel(dq, dq_def)
    mmap = {
        b'FOO': 4,
        b'BAR': 1,
        b'BAM': 2,
    }
    dqmask = dynamic_mask(model, mmap, inv)
    if inv:
        numpy.testing.assert_equal(dqmask, [0, 2, 4, 1])
    else:
        numpy.testing.assert_equal(dqmask, [0, 4, 1, 2])


@pytest.mark.parametrize('inv', [True, False])
def test_missing_map_key(caplog, inv):
    dq = _dq([0, 1])
    dq_def = _dq_def([(1, 'FOO')])
    model = FakeModel(dq, dq_def)
    mmap = {
        b'BAR': 2,
    }
    dqmask = dynamic_mask(model, mmap, inv)
    numpy.testing.assert_equal(dqmask, [0, 0])
    assert 'does not correspond to an existing DQ' in caplog.text
